
try:
    from pandas_reader import PandasReader
    from plot_builder import PlotBuilder
    from float_list_helper import FloatListHelper
except ImportError:
    from .pandas_reader import PandasReader
    from .plot_builder import PlotBuilder
    from .float_list_helper import FloatListHelper
    
import os

class DtaProcessor:
    def __init__(self, csv_file_path, model_colunm_index, 
            experimental_data_column_index, output_dir_path):
        self.csv_file_path = csv_file_path
        self.model_colunm_index = model_colunm_index
        self.experimental_data_column_index = experimental_data_column_index

        self.model_colunm_index = model_colunm_index
        self.experimental_data_column_index = experimental_data_column_index
        self.output_dir_path = output_dir_path


    def process(self):
        # Get model and difference lists
        model_list, difference_list = self.get_model_and_difference_lists()

        PlotBuilder.get_graph_image(
            model_list,
            difference_list,
            "Model vs Experimental Data",
            "Model",
            "Difference",
            os.path.join(self.output_dir_path, "model_vs_experimental_data.png")
        )

        line_1_coords = DtaProcessor.get_line_1_coords(model_list, difference_list)
        line_2_coords = DtaProcessor.get_line_2_coords(model_list, difference_list)

        PlotBuilder.get_graph_image_with_lines(
            model_list,
            difference_list,
            "Model vs Experimental Data",
            "Model",
            "Difference",
            os.path.join(self.output_dir_path, "model_vs_experimental_data_with_lines.png"),
            [line_1_coords, line_2_coords]
        )

        print("Line 1 Coords:", line_1_coords)
        print("Line 2 Coords:", line_2_coords)

        intersection_point = DtaProcessor.find_intersection(line_1_coords, line_2_coords)
        print("Intersection Point:", intersection_point)


    def get_line_1_coords(model_list, difference_list):
        max_index, max_val = FloatListHelper.find_max_index(model_list)

        model_list_cropped, difference_list_cropped = \
            FloatListHelper.crop_lists(
                model_list, 
                difference_list, 
                max_index
            )

        start_index = FloatListHelper.find_start_index(model_list_cropped, difference_list_cropped)
        end_index = FloatListHelper.find_end_index(model_list_cropped, difference_list_cropped)
        line_1_coords = {
            "x_0": model_list_cropped[start_index], 
            "y_0": difference_list_cropped[start_index], 
            "x_1": model_list_cropped[end_index], 
            "y_1": difference_list_cropped[end_index]
        }

        return line_1_coords


    def get_line_2_coords(model_list, difference_list):
        max_index, max_val = FloatListHelper.find_max_index(model_list)

        model_list_cropped, difference_list_cropped = \
            FloatListHelper.crop_lists(
                model_list, 
                difference_list, 
                max_index
            )
        
        start_index = FloatListHelper.find_start_index(model_list_cropped, difference_list_cropped, 0.3)
        difference_list_cropped_15 = difference_list_cropped[start_index:]
        model_list_cropped_15 = model_list_cropped[start_index:]
        
        max_index, max_val = FloatListHelper.find_max_index(difference_list_cropped_15)

        difference_list_cropped_2 = difference_list_cropped[start_index:start_index + max_index]
        min_index, min_val = FloatListHelper.find_min_index(difference_list_cropped_2)

        line_2_coords = {
            "x_0": model_list_cropped[start_index + min_index],
            "y_0": difference_list_cropped[start_index + min_index],
            "x_1": model_list_cropped_15[max_index],
            "y_1": difference_list_cropped_15[max_index]
        }

        return line_2_coords
    

    def find_intersection(line1_coords, line2_coords):
        """
        Find the intersection point of two lines.

        Args:
            line1_coords (dict): Dictionary with coordinates of the first line.
            line2_coords (dict): Dictionary with coordinates of the second line.

        Returns:
            tuple: Intersection point coordinates.

        """
        x1, y1 = line1_coords["x_0"], line1_coords["y_0"]
        x2, y2 = line1_coords["x_1"], line1_coords["y_1"]
        x3, y3 = line2_coords["x_0"], line2_coords["y_0"]
        x4, y4 = line2_coords["x_1"], line2_coords["y_1"]

        A1 = y2 - y1
        B1 = x1 - x2
        C1 = A1 * x1 + B1 * y1

        A2 = y4 - y3
        B2 = x3 - x4
        C2 = A2 * x3 + B2 * y3

        determinant = A1 * B2 - A2 * B1

        if determinant == 0:
            return None
        else:
            x = (B2 * C1 - B1 * C2) / determinant
            y = (A1 * C2 - A2 * C1) / determinant
            return (x, y)
        

    def get_model_and_difference_lists(self):
        """
        Get model and difference lists from the csv file.
        
        Returns:
            tuple: Model and difference lists.
            
        """
        pandas_reader = PandasReader(self.csv_file_path)
        pandas_data = pandas_reader.read()

        graph_data = PandasReader.extract_columns(
            pandas_data, 
            [self.model_colunm_index, self.experimental_data_column_index]
        )
        
        PandasReader.check_for_strings(graph_data)

        graph_data_2 = PandasReader.convert_strings_to_floats(graph_data)

        PandasReader.check_for_strings(graph_data_2)

        # Substract model column from experimental data column, adding new column to the data frame
        # and plot the graph
        
        try:
            difference = graph_data_2[self.experimental_data_column_index] - graph_data_2[self.model_colunm_index]
        except KeyError as e:
            print(f"KeyError: {e}")
            # Handle the error or re-raise
            raise

        difference_list = difference.values.tolist()
        model_list = graph_data_2[self.model_colunm_index].values.tolist()

        return model_list, difference_list


import matplotlib.pyplot as plt

class PlotBuilder:

    @staticmethod
    def get_graph_image(x, y, title, x_label, y_label, output_path):
        """
        Create a graph.

        Args:
            x (list): List of x values.
            y (list): List of y values.
            title (str): Title of the graph.
            x_label (str): Label for the x axis.
            y_label (str): Label for the y axis.
            output_path (str): Path to save the graph image.

        """
        plt.plot(x, y)
        plt.title(title)
        plt.xlabel(x_label)
        plt.ylabel(y_label)
        plt.grid()
        plt.savefig(output_path)
        plt.close()

    
    @staticmethod
    def get_graph_image_with_lines(x, y, title, x_label, y_label, output_path, lines_coords):
        """
        Create a graph with lines.
        
        Args:
            x (list): List of x values.
            y (list): List of y values.
            title (str): Title of the graph.
            x_label (str): Label for the x axis.
            y_label (str): Label for the y axis.
            output_path (str): Path to save the graph image.
            lines_coords (list[dict]): List of dictionaries with line coordinates.

        """
        plt.plot(x, y)
        plt.title(title)
        plt.xlabel(x_label)
        plt.ylabel(y_label)
        plt.grid()
        colors = ['r', 'g', 'b', 'y', 'm', 'c']
        for line_coords in lines_coords:
            plt.axline((line_coords["x_0"], line_coords["y_0"]), (line_coords["x_1"], line_coords["y_1"]), color=colors.pop(0))
        plt.savefig(output_path)
        plt.close()


import sys
import os

sys.path.append(os.path.join(os.path.dirname(__file__), '..'))

from dtalib import DtaProcessor

def main():
    dta_processor = DtaProcessor(
        csv_file_path="data/test_data_5.csv", 
        model_colunm_index=17,
        experimental_data_column_index=16,
        output_dir_path="data"
    )

    dta_processor.process()

if __name__ == "__main__":
    main()




class FloatListHelper:
    @staticmethod
    def crop_lists(x, y, start_index):
        """
        Crop lists to the same length starting from the specified index.

        Args:
            x (list): List of x values.
            y (list): List of y values.
            start (int): Index to start cropping from.

        Returns:
            tuple: Cropped lists.

        """
        x_cropped = x[start_index:]
        y_cropped = y[start_index:]

        return x_cropped, y_cropped
    
    def find_max_index(lst):
        """
        Find the index of the maximum value in the list.

        Args:
            lst (list): List of float values.

        Returns:
            int: Index of the maximum value.
        """
        return lst.index(max(lst)), max(lst)
    

    def find_min_index(lst):
        """
        Find the index of the minimum value in the list.

        Args:
            lst (list): List of float values.

        Returns:
            int: Index of the minimum value.
        """
        return lst.index(min(lst)), min(lst)
    

    @staticmethod
    def find_end_index(x, y):
        """
        Find the index of the end of the list.

        Args:
            x (list): List of x values.
            y (list): List of y values.

        Returns:
            int: Index of the end of the list.
        """
        return len(x) - 1
    
    
    def find_start_index(x_cropped, y_cropped, coefficient=0.75):
        """
        Find the index of the start of the list.

        Args:
            x_cropped (list): List of x values.
            y_cropped (list): List of y values.
            coefficient (float): Coefficient to estimate the start index.

        Returns:
            int: Index of the start of the list.

        """
        
        min_x = min(x_cropped)
        max_x = max(x_cropped)

        estimated_start = min_x + (max_x - min_x) * coefficient

        start_index, start_val = FloatListHelper.find_closest_index(x_cropped, estimated_start)

        return start_index


    def find_closest_index(lst, value):
        """
        Find the index of the closest value in the list.

        Args:
            lst (list): List of float values.
            value (float): Value to find.

        Returns:
            int: Index of the closest value.

        """
        return min(range(len(lst)), key=lambda i: abs(lst[i] - value)), min(lst, key=lambda x:abs(x-value))


import pandas as pd

class PandasReader:
    def __init__(self, file_path):
        self.file_path = file_path
        self.encoding = 'utf-8'
        self.windows_encoding = 'windows-1251'


    def read(self):
        """
        Reads a CSV file and returns a DataFrame.
        
        Returns:
            pd.DataFrame: DataFrame with the data from the CSV file.

        """
        rows_to_skip = 5
        try:
            return pd.read_csv(
                self.file_path, 
                encoding=self.encoding, 
                header=None,
                skiprows=rows_to_skip,
                skipfooter=1,
            )
        except UnicodeDecodeError:
            print("Error: utf-8 encoding not working. Trying windows-1251 encoding")
            # skip the first row
            return pd.read_csv(
                self.file_path, 
                encoding=self.windows_encoding,
                header=None,
                skiprows=rows_to_skip,
                skipfooter=1,
            )
        except Exception as e:
            return str(e)
        

    @staticmethod
    def count_numbers(row):
        """
        Counts the number of non-null values in a row.

        Args:
            row (pd.Series): Row of a DataFrame.

        Returns:
            int: Number of non-null values.
        """
        count = 0
        print(row)
        for val in row:
            print(val)
            if not pd.isnull(val):
                count += 1
        print(count)
        return count


    def find_essential_start(self, df, n):
        """
        Finds the index of the first row with at least n non-null values.
        
        Args:
            df (pd.DataFrame): DataFrame with the data.
            n (int): Minimum number of non-null values.

        Returns:
            int: Index of the first row with at least n non-null values.

        """
        return 0


    def extract_essential(self, df, start_index):
        """
        Extracts rows starting from the specified index.

        Args:
            df (pd.DataFrame): DataFrame with the data.
            start_index (int): Index to start extracting from.

        Returns:
            pd.DataFrame: DataFrame with the extracted rows.

        """
        return df[start_index:] if df is not None else None


    @staticmethod
    def extract_columns(df, col_indices):
        """
        Extracts specified columns from a DataFrame.

        Args:
            df (pd.DataFrame): DataFrame with the data.
            col_indices (list): List of column indices to extract.

        Returns:
            pd.DataFrame: DataFrame with the extracted columns.

        """
        # Разбиваем строки на отдельные колонки по разделителю
        split_df = df.iloc[:, 0].str.split(';', expand=True)
        
        # Извлекаем указанные колонки
        extracted_df = split_df.iloc[:, col_indices]
        
        return extracted_df
    

    @staticmethod
    def substract_columns(df, col1, col2):
        """
        Subtracts values of two columns in a DataFrame.

        """
        new_column = df[col1] - df[col2]
        return new_column
    
    @staticmethod
    def dataframe_to_list(df):
        """
        Converts a DataFrame to a list of lists.

        Args:
            df (pd.DataFrame): DataFrame to convert.

        Returns:
            list: List of lists.
        """
        return df.values.tolist()
    

    @staticmethod
    def check_for_strings(df):
        """
        Checks for string values in a DataFrame and prints them.

        Args:
            df (pd.DataFrame): DataFrame

        """
        for column in df.columns:
            string_values = df[df[column].apply(lambda x: isinstance(x, str))]
            if not string_values.empty:
                pass


    @staticmethod
    def convert_strings_to_floats(df):
        """
        Converts string values in a DataFrame to floats.

        Args:
            df (pd.DataFrame): DataFrame

        Returns:
            pd.DataFrame: DataFrame with converted values.

        """
        return df.apply(pd.to_numeric, errors='coerce')




__version__ = "1.0.22"

from .plot_builder import PlotBuilder
from .pandas_reader import PandasReader
from .dta_processor import DtaProcessor

__all__ = [
    "PlotBuilder",
    "PandasReader",
    "DtaProcessor"
]